"""
import_summary.py
Update the summary spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import read_piece_definitions, read_template
from ._output_files import (
    SI_SUMMARY_KEY,
    read_spreadsheets_index,
    read_summary,
    write_spreadsheets_index,
)
from ._sheets import (
    add_permissions,
    add_temp_sheet,
    create_spreadsheet,
    gspread_auth,
    open_spreadsheet,
)

# ======================================================================

__all__ = ("import_summary",)

# ======================================================================


def _create_piece_sheets(spreadsheet, pieces, summary):
    """Create the piece sheets in the summary spreadsheet.

    Args:
        spreadsheet (gspread.Spreadsheet): The summary spreadsheet.
        pieces (Dict[str, Piece]): A mapping from piece names to piece
            objects.
        summary (Dict[str, Dict]): The piece summary.

    Returns:
        bool: Whether the creations were successful.
    """

    # delete existing sheets
    existing_sheets = list(spreadsheet.worksheets())
    success, temp_sheet = add_temp_sheet(spreadsheet, invalid=summary.keys())
    if not success:
        return False
    for sheet in existing_sheets:
        spreadsheet.del_worksheet(sheet)

    for title in summary.keys():
        logger.debug('Creating sheet for piece "{}"', title)
        pieces[title].create_summary_sheet(spreadsheet, summary[title])

    # delete temp sheet
    spreadsheet.del_worksheet(temp_sheet)

    return True


# ======================================================================


@click.command("import_summary", help="Update the summary spreadsheet.")
@click.option(
    "-c",
    "--create",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Create a new summary spreadsheet.",
)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-s",
    "summary_path",
    type=str,
    help="A filepath to replace the summary file.",
)
@click.option(
    "-si", type=str, help="A filepath to replace the spreadsheets index file."
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def import_summary(create, td, pd, summary_path, si, strict):
    """Update the summary spreadsheet.

    Args:
        create (str): Whether to create a new summary spreadsheet.
            Default is False.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        summary_path (str): A filepath to replace the summary file.
        si (str): A filepath to replace the spreadsheets index file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    logger.warning(
        "For most accurate summary, run the `compile_pieces` or "
        "`export_summary` command first."
    )

    success, gc = gspread_auth()
    if not success:
        return

    success, template = read_template(td, strict)
    if not success:
        return

    success, (pieces, pieces_data) = read_piece_definitions(
        template, pd, as_both=True
    )
    if not success:
        return

    success, summary = read_summary(pieces_data, summary_path, strict)
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return

    # check if need to create spreadsheet
    if SI_SUMMARY_KEY not in spreadsheets:
        logger.debug(
            "Summary spreadsheet link not found in spreadsheets index "
            "file: creating new"
        )
        create = True

    # create or open spreadsheet
    if create:
        ss_settings = template["summarySpreadsheet"]

        success, spreadsheet = create_spreadsheet(
            gc, ss_settings["title"], ss_settings["folder"]
        )
        if not success:
            return

        # permissions
        success = add_permissions(
            spreadsheet, ss_settings["publicAccess"], ss_settings["shareWith"]
        )
        if not success:
            gc.del_spreadsheet(spreadsheet.id)
            return

        # save link
        spreadsheets[SI_SUMMARY_KEY] = spreadsheet.url
    else:
        link = spreadsheets[SI_SUMMARY_KEY]
        success, spreadsheet = open_spreadsheet(gc, link)
        if not success:
            return

    # create pieces
    success = _create_piece_sheets(spreadsheet, pieces, summary)
    if create and not success:
        # delete created spreadsheet
        gc.del_spreadsheet(spreadsheet.id)
        return

    if create:
        success = write_spreadsheets_index(spreadsheets, si)
        if not success:
            return

    logger.info("Done")



"""
compile_pieces.py
Compile all piece JSON data flies into a single file.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import read_piece_definitions, read_template
from ._output_files import read_piece_data, write_summary
from ._shared import error

# ======================================================================

__all__ = ("compile_pieces",)

# ======================================================================


@click.command(
    "compile_pieces",
    help="Compile all piece JSON data flies into a single file.",
)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-pdp",
    type=str,
    help=(
        "A filepath to replace the piece data path file. "
        'Must include "{piece}".'
    ),
)
@click.option(
    "-s",
    "summary_path",
    type=str,
    help="A filepath to replace the summary file.",
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def compile_pieces(td, pd, pdp, summary_path, strict):
    """Compile all piece JSON data flies into a single file.

    Args:
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        pdp (str): A filepath to replace the piece data path file.
            Must include "{piece}" exactly once.
        summary_path (str): A filepath to replace the summary file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    # validate args
    if pdp is not None and pdp.count("{piece}") != 1:
        error('`pdp` must include "{piece}" exactly once')
        return

    logger.warning(
        "For most accurate summary, run the `piece_summary` command first."
    )

    success, template = read_template(td, strict)
    if not success:
        return

    success, pieces = read_piece_definitions(template, pd, as_data=True)
    if not success:
        return

    success, pieces_data = read_piece_data(pieces, pdp, strict)
    if not success:
        return

    summary = []
    for title, piece in pieces_data.items():
        piece_obj = pieces[title]
        # turn sources into list
        piece["sources"] = list(piece["sources"].values())
        for source in piece["sources"]:
            source["summary"] = piece_obj.make_default()
        summary.append(piece)

    success = write_summary(summary, summary_path)
    if not success:
        return

    logger.info("Done")



"""
__main__.py
The command-line interface for Roseingrave.
"""

# ======================================================================

import click

from .compile_pieces import compile_pieces
from .create_sheet import create_sheet
from .export_summary import export_summary
from .fix_input import fix_input
from .import_summary import import_summary
from .piece_summary import piece_summary
from .reauth import reauth
from .volunteer_summary import volunteer_summary

# ======================================================================


class OrderedGroup(click.Group):
    """Lists the commands in the listed order (instead of alphabetical)."""

    def __init__(self, *args, **kwargs):
        commands = kwargs.get("commands", None)
        if commands is not None:
            if isinstance(commands, list):
                kwargs["commands"] = {cmd.name: cmd for cmd in commands}
        super().__init__(*args, **kwargs)

    def list_commands(self, ctx):
        return list(self.commands.keys())


cli = OrderedGroup(
    commands=[
        reauth,
        fix_input,
        create_sheet,
        volunteer_summary,
        piece_summary,
        compile_pieces,
        import_summary,
        export_summary,
    ],
    help="Massively scalable musical source comparator",
)

if __name__ == "__main__":
    cli()



"""
volunteer_summary.py
Export volunteer JSON data files.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import (
    read_piece_definitions,
    read_template,
    read_volunteer_definitions,
)
from ._output_files import (
    SI_SUMMARY_KEY,
    read_spreadsheets_index,
    write_volunteer_data,
)
from ._piece import Piece
from ._shared import error, fail_on_warning
from ._sheets import gspread_auth, open_spreadsheet

# ======================================================================

__all__ = ("volunteer_summary",)

# ======================================================================


def export_spreadsheet(gc, email, link, template, volunteer=None):
    """Export a volunteer's spreadsheet.

    Args:
        gc (gspread.Client): The client.
        email (str): The volunteer email.
        link (str): The spreadsheet link.
        template (Dict): The template settings.
        volunteer (Optional[Volunteer]): The volunteer. If given, only
            export the pieces belonging to this volunteer.
            Default is None.

    Returns:
        Tuple[bool, List[Dict]]: Whether the export was successful,
            and the exported data in JSON format.
    """
    ERROR_RETURN = False, None

    logger.debug('Working on volunteer "{}"', email)

    volunteer_pieces = None
    if volunteer is not None:
        volunteer_pieces = set(volunteer.pieces)

    success, spreadsheet = open_spreadsheet(gc, link)
    if not success:
        return ERROR_RETURN

    data = []
    for sheet in spreadsheet.worksheets():
        if volunteer_pieces is not None:
            if sheet.title not in volunteer_pieces:
                logger.warning(
                    'Volunteer "{}" does not have piece "{}" in piece '
                    "definitions file: skipping",
                    email,
                    sheet.title,
                )
                continue
        success, piece_data = Piece.export_sheet(sheet, template)
        if not success:
            continue
        data.append(piece_data)

    return True, data


# ======================================================================


@click.command(
    "volunteer_summary",
    help="Export volunteer JSON data files.",
)
@click.argument("emails", type=str, nargs=-1)
@click.option(
    "-si", type=str, help="A filepath to replace the spreadsheets index file."
)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-vd",
    type=str,
    help="A filepath to replace the volunteer definitions file.",
)
@click.option(
    "-vdp",
    type=str,
    help=(
        "A filepath to replace the volunteer data path file. "
        'Must include "{email}".'
    ),
)
@click.option(
    "-ek",
    "--export-known-only",
    is_flag=True,
    default=False,
    flag_value=True,
    help=(
        "Export only the volunteers and pieces that appear in the "
        "definition files. Requires piece definitions file and "
        "volunteer definitions file. Default is False."
    ),
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def volunteer_summary(emails, si, td, pd, vd, vdp, export_known_only, strict):
    """Export volunteer JSON data files.

    Args:
        emails (Tuple[str, ...]): The volunteers to export data for. If
            none given, exports data for all volunteers.
        si (str): A filepath to replace the spreadsheets index file.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        vd (str): A filepath to replace the volunteer definitions file.
        vdp (str): A filepath to replace the volunteer data path file.
            Must include "{email}" exactly once.
        export_known_only (bool): Whether to export only the volunteers
            and pieces that appear in the definition files.
            Requires piece definitions file and volunteer definitions
            file.
            Default is False.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    # validate args
    if vdp is not None and vdp.count("{email}") != 1:
        error('`vdp` must include "{email}" exactly once')
        return

    success, gc = gspread_auth()
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si, must_exist=True)
    if not success:
        return
    # remove summary spreadsheet
    spreadsheets.pop(SI_SUMMARY_KEY, None)

    success, template = read_template(td, strict)
    if not success:
        return

    if len(emails) > 0:
        # filter volunteers in `emails`
        filtered = {}
        for email in emails:
            if email not in spreadsheets:
                logger.warning(
                    'Volunteer "{}" not found in spreadsheets index file',
                    email,
                )
                continue
            filtered[email] = spreadsheets[email]
        spreadsheets = filtered

    if len(spreadsheets) == 0:
        logger.info("No volunteers to export data for")
        return

    if export_known_only:
        success, pieces = read_piece_definitions(template, pd)
        if not success:
            return

        success, volunteers = read_volunteer_definitions(pieces, vd, strict)
        if not success:
            return
    else:
        volunteers = None

    # volunteer email -> data
    data = {}

    # export spreadsheets
    logger.info("Exporting data from spreadsheets")
    for email, link in spreadsheets.items():
        volunteer = None
        if volunteers is not None:
            if email not in volunteers:
                logger.warning(
                    'Unknown volunteer "{}": skipping spreadsheet', email
                )
                continue
            volunteer = volunteers[email]
        success, volunteer_data = export_spreadsheet(
            gc, email, link, template, volunteer
        )
        if strict and not success:
            fail_on_warning()
            return
        if success:
            data[email] = volunteer_data

    if len(data) > 0:
        success = write_volunteer_data(data, vdp)
        if not success:
            return

    logger.info("Done")



"""
piece_summary.py
Export piece JSON data files.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import read_piece_definitions, read_template
from ._output_files import read_volunteer_data, write_piece_data
from ._shared import error

# ======================================================================

__all__ = ("piece_summary",)

# ======================================================================


def _extract_pieces(pieces, volunteers_data, strict):
    """Extract pieces from volunteer data files.

    Args:
        pieces (Dict[str, PieceData]): A mapping from piece names to
            piece objects.
            Only includes the pieces needed.
        volunteers_data (Dict[str, List]): A mapping from volunteer
            emails to data.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.

    Returns:
        Tuple[bool, Dict[str, Dict]]:
            Whether the extraction was successful, and
            a mapping from piece names to data.
    """
    ERROR_RETURN = False, None

    logger.info("Extracting pieces from volunteer data files")

    seen = {title: False for title in pieces.keys()}

    for email, volunteer_data in volunteers_data.items():
        for title, piece in volunteer_data.items():
            # not being extracted, so ignore
            if title not in pieces:
                continue
            seen[title] = True
            success = pieces[title].add_volunteer(email, piece, strict)
            if not success:
                return ERROR_RETURN

    # convert PieceData to dicts
    # only include the ones actually seen
    data = {
        title: piece.to_json()
        for title, piece in pieces.items()
        if seen[title]
    }

    if len(data) == 0:
        logger.info("No data found for pieces")
        return ERROR_RETURN

    return True, data


# ======================================================================


@click.command(
    "piece_summary",
    help="Export piece JSON data files.",
)
@click.argument("pieces", type=str, nargs=-1)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-vdp",
    type=str,
    help=(
        "A filepath to replace the volunteer data path file. "
        'Must include "{email}".'
    ),
)
@click.option(
    "-pdp",
    type=str,
    help=(
        "A filepath to replace the piece data path file. "
        'Must include "{piece}".'
    ),
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def piece_summary(pieces, td, pd, vdp, pdp, strict):
    """Export piece JSON data files.

    Args:
        pieces (Tuple[str, ...]): The pieces to export data for. If none
            given, exports data for all pieces found.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        vdp (str): A filepath to replace the volunteer data path file.
            Must include "{email}" exactly once.
        pdp (str): A filepath to replace the piece data path file.
            Must include "{piece}" exactly once.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    # validate args
    if vdp is not None and vdp.count("{email}") != 1:
        error('`vdp` must include "{email}" exactly once')
        return
    if pdp is not None and pdp.count("{piece}") != 1:
        error('`pdp` must include "{piece}" exactly once')
        return

    logger.warning(
        "For most accurate summary, run the `volunteer_summary` "
        "command first."
    )

    success, template = read_template(td, strict)
    if not success:
        return

    success, piece_data = read_piece_definitions(template, pd, as_data=True)
    if not success:
        return

    success, volunteers_data = read_volunteer_data(piece_data, vdp, strict)
    if not success:
        return

    # filter the pieces to extract
    if len(pieces) > 0:
        filtered = {}
        for piece in pieces:
            if piece not in piece_data:
                logger.warning(
                    'Piece "{}" not found in piece definitions file', piece
                )
                continue
            filtered[piece] = piece_data[piece]
        piece_data = filtered

    if len(piece_data) == 0:
        logger.info("No pieces to summarize data for")
        return

    success, data = _extract_pieces(piece_data, volunteers_data, strict)
    if not success:
        return

    for piece in pieces:
        if piece not in data:
            logger.warning('No data found for piece "{}"', piece)

    success = write_piece_data(data, pdp)
    if not success:
        return

    logger.info("Done")



"""
fix_input.py
Fixes input files.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import (
    fix_piece_definitions,
    fix_volunteer_definitions,
    read_piece_definitions,
    read_template,
)
from ._output_files import fix_spreadsheets_index
from ._read_write import fix_settings
from ._shared import error

# ======================================================================

__all__ = ("fix_input",)

# ======================================================================

FILES = {
    key: True
    for key in (
        "settings",
        "pieces",
        "volunteers",
        "spreadsheetsIndex",
    )
}

# ======================================================================


def _fix_volunteers(pieces_error, pieces, pd=None, vd=None):
    """Fix the volunteer definitions file.

    Args:
        pieces_error (bool): Whether there was an error fixing the
            piece definitions file.
        pieces (Dict[str, Piece]): The known pieces.
        pd (Optional[str]): A filepath to replace the piece definitions
            file.
        vd (Optional[str]): A filepath to replace the volunteer
            definitions file.

    Returns:
        bool: Whether the fix was successful.
    """
    ERROR_RETURN = False, None

    if pieces_error:
        error('Pieces file could not be read to fix "volunteers"')
        return ERROR_RETURN

    # read the piece definitions (don't fix)
    if pieces is None:
        success, template = read_template()
        if not success:
            return ERROR_RETURN
        success, pieces = read_piece_definitions(template, pd)
        if not success:
            return ERROR_RETURN

    return fix_volunteer_definitions(pieces, vd)


# ======================================================================


@click.command("fix_input", help="Fixes input files.")
@click.argument("files", type=str, nargs=-1)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-vd",
    type=str,
    help="A filepath to replace the volunteer definitions file.",
)
@click.option(
    "-si", type=str, help="A filepath to replace the spreadsheets index file."
)
def fix_input(files, pd, vd, si):
    """Fixes input files.

    Args:
        files (Tuple[str, ...]): The files to fix.
            If none given, all files will be fixed.
        pd (str): A filepath to replace the piece definitions file.
        vd (str): A filepath to replace the volunteer definitions file.
        si (str): A filepath to replace the spreadsheets index file.
    """

    if len(files) == 0:
        files = set(FILES.keys())
    else:
        # filter out unknown files
        filtered = set()
        for file in files:
            if file not in FILES:
                logger.warning('Unknown file key "{}"', file)
                continue
            filtered.add(file)
        files = filtered

    if len(files) == 0:
        logger.info("No files to fix")
        return

    # fix settings
    if "settings" in files:
        _ = fix_settings()

    # fix pieces
    pieces_error, pieces = False, None
    if "pieces" in files:
        pieces_error, pieces = fix_piece_definitions(pd)

    # fix volunteers
    if "volunteers" in files:
        _ = _fix_volunteers(pieces_error, pieces, pd, vd)

    # fix spreadsheets
    if "spreadsheetsIndex" in files:
        _ = fix_spreadsheets_index(si)



"""
_volunteer.py
The Volunteer class.
"""

# ======================================================================

__all__ = ("Volunteer",)

# ======================================================================


class Volunteer:
    """Volunteer class."""

    def __init__(self, kwargs, known_pieces):
        """Initialize a volunteer from a JSON dict.

        Args:
            kwargs (Dict): The JSON dict.
            known_pieces (Dict[str, Piece]): The known pieces.

        Raises:
            ValueError: If any required key is not found.
        """

        for key in ("email", "pieces"):
            if key not in kwargs:
                raise ValueError(f'key "{key}" not found')

        self._email = kwargs["email"]
        self._pieces = {}
        self._unknown = []
        for piece_name in kwargs["pieces"]:
            if piece_name in self._pieces:
                continue
            piece = known_pieces.get(piece_name, None)
            if piece is not None:
                self._pieces[piece_name] = True
            else:
                self._unknown.append(piece_name)

    @property
    def email(self):
        return self._email

    @property
    def pieces(self):
        return list(self._pieces.keys())

    @property
    def unknown_pieces(self):
        return self._unknown

    def combine(self, other):
        """Combine this volunteer with another by taking the union of
        all pieces.
        """
        for piece_name in other.pieces:
            if piece_name not in self._pieces:
                self._pieces[piece_name] = True

    def to_json(self):
        """Return a JSON representation of this volunteer."""
        return {
            "email": self._email,
            "pieces": list(self._pieces.keys()),
        }



"""
export_summary.py
Export the summary spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import (
    read_piece_definitions,
    read_template,
    read_volunteer_definitions,
)
from ._output_files import (
    SI_SUMMARY_KEY,
    read_spreadsheets_index,
    write_summary,
)
from ._piece import Piece
from ._shared import error, fail_on_warning
from ._sheets import gspread_auth, open_spreadsheet

# ======================================================================

__all__ = ("export_summary",)

# ======================================================================


def filter_volunteers(piece_name, piece_volunteers, piece_data):
    removed = set()

    def remove_volunteers(obj):
        for volunteer in list(obj.keys()):
            if volunteer in piece_volunteers:
                continue
            if volunteer not in removed:
                # only log warning once per unknown volunteer
                logger.warning(
                    'Unknown volunteer "{}" for piece "{}" (not in '
                    "volunteer definitions file)",
                    volunteer,
                    piece_name,
                )
                removed.add(volunteer)
            obj.pop(volunteer)

    # remove from sources
    for source in piece_data["sources"]:
        remove_volunteers(source["volunteers"])
        # can't do anything about "summary", since that's inputted from
        # the user
    # remove from notes
    for key, volunteer_data in piece_data["notes"].items():
        if key == "bars":
            # special case: bars
            for bar_data in volunteer_data.values():
                remove_volunteers(bar_data)
        else:
            remove_volunteers(volunteer_data)


# ======================================================================


@click.command("export_summary", help="Export the summary spreadsheet.")
@click.option(
    "-si", type=str, help="A filepath to replace the spreadsheets index file."
)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-vd",
    type=str,
    help="A filepath to replace the volunteer definitions file.",
)
@click.option(
    "-s",
    "summary_path",
    type=str,
    help="A filepath to replace the summary file.",
)
@click.option(
    "-ek",
    "--export-known-only",
    is_flag=True,
    default=False,
    flag_value=True,
    help=(
        "Export only the volunteers and pieces that appear in the "
        "definition files. Requires piece definitions file and "
        "volunteer definitions file. Default is False."
    ),
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def export_summary(si, td, pd, vd, summary_path, export_known_only, strict):
    """Export the summary spreadsheet.

    Args:
        si (str): A filepath to replace the spreadsheets index file.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        vd (str): A filepath to replace the volunteer definitions file.
        summary_path (str): A filepath to replace the summary file.
        export_known_only (bool): Whether to export only the volunteers
            and pieces that appear in the definition files.
            Requires piece definitions file and volunteer definitions
            file.
            Default is False.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """

    success, gc = gspread_auth()
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si, must_exist=True)
    if not success:
        return

    if SI_SUMMARY_KEY not in spreadsheets:
        error(
            f'Summary spreadsheet (key "{SI_SUMMARY_KEY}") not found in '
            "spreadsheets index file"
        )
        return

    success, template = read_template(td, strict)
    if not success:
        return

    if export_known_only:
        success, pieces = read_piece_definitions(template, pd)
        if not success:
            return

        success, volunteers = read_volunteer_definitions(pieces, vd, strict)
        if not success:
            return

        piece_volunteers = {piece_name: set() for piece_name in pieces.keys()}
        for email, volunteer in volunteers.items():
            # the volunteer's pieces definitely exist
            for piece_name in volunteer.pieces:
                piece_volunteers[piece_name].add(email)
    else:
        piece_volunteers = None

    success, spreadsheet = open_spreadsheet(gc, spreadsheets[SI_SUMMARY_KEY])
    if not success:
        return

    summary = []
    for sheet in spreadsheet.worksheets():
        if piece_volunteers is not None:
            if sheet.title not in piece_volunteers:
                logger.warning(
                    'Unknown piece "{}" (not in piece definitions '
                    "file: skipping sheet",
                    sheet.title,
                )
                continue
        success, piece_data = Piece.export_summary_sheet(sheet, template)
        if not success:
            if strict:
                fail_on_warning()
                return
            continue
        if piece_volunteers is not None:
            filter_volunteers(
                sheet.title, piece_volunteers[sheet.title], piece_data
            )
        summary.append(piece_data)

    if len(summary) > 0:
        success = write_summary(summary, summary_path)
        if not success:
            return

    logger.info("Done")



"""
_piece_data.py
The PieceData and SourceData classes.
"""

# ======================================================================

from loguru import logger

from ._shared import fail_on_warning

# ======================================================================

__all__ = ("PieceData",)

# ======================================================================


class SourceData:
    """The data of a source from data files."""

    def __init__(self, piece, source):
        """Initialize the source data from a source.

        Args:
            piece (PieceData): The parent piece.
            source (Source): The source.
        """

        self._piece = piece

        self._name = source.name
        self._link = source.link
        self._volunteers = {}

    @property
    def name(self):
        return self._name

    @property
    def link(self):
        return self._link

    def to_json(self, has_summary=False):
        """Return a JSON representation of this source data.

        Args:
            has_summary (bool): Whether this source data should contain
                a "summary" volunteer.
                Default is False.

        Returns:
            Dict: A JSON dict.
        """
        volunteers = self._volunteers
        if has_summary:
            volunteers["summary"] = self._piece.make_default()
        return {
            "name": self._name,
            "link": self._link,
            "volunteers": volunteers,
        }

    def add_volunteer(self, email, data):
        """Add a volunteer's data to this source.

        Args:
            email (str): The volunteer email.
            data (Dict): The volunteer data.

        Returns:
            bool: Whether the addition was successful.
        """
        self._volunteers[email] = {
            key: val
            for key, val in data.items()
            if key not in ("name", "link")
        }


class PieceData:
    """The data of a piece from data files."""

    def __init__(self, piece):
        """Initialize the piece data from a piece.

        Args:
            piece (Piece): The piece.
        """

        self._bar_count = piece.final_bar_count
        self._headers = list(piece._template["metaDataFields"].keys())

        self._name = piece.name
        self._link = piece.link
        self._sources = {
            source.name: SourceData(self, source) for source in piece.sources
        }
        self._notes = self.make_default(True)

    @property
    def name(self):
        return self._name

    @property
    def link(self):
        return self._link

    @property
    def bar_count(self):
        return self._bar_count

    def all_sources(self):
        """Return list of all source names."""
        return list(self._sources.keys())

    def has_source(self, name):
        """Check if this piece has a source with the given name."""
        return name in self._sources

    def get_source(self, name):
        """Get the source by the given name, or None."""
        return self._sources.get(name, None)

    def to_json(self, has_summary=False):
        """Return a JSON representation of this piece data.

        Args:
            has_summary (bool): Whether the source data should include
                a "summary" volunteer.
                Default is False.

        Returns:
            Dict: A JSON dict.
        """
        return {
            "title": self._name,
            "link": self._link,
            "sources": [
                source.to_json(has_summary)
                for source in self._sources.values()
            ],
            "notes": self._notes,
        }

    def make_default(self, is_notes=False):
        """Make a JSON dict with the default values."""

        if is_notes:

            def value():
                return {}

        else:

            def value():
                return ""

        # template names
        values = {key: value() for key in self._headers}
        # bars
        values["bars"] = {
            str(bar_num + 1): value() for bar_num in range(self._bar_count)
        }
        # comments
        if not is_notes:
            values["comments"] = value()

        return values

    def with_defaults(
        self, values, loc, exclude_comments=False, is_notes=False
    ):
        """Fix a JSON dict by adding the default values.
        Displays warnings for missing or extra fields.

        Args:
            values (Dict): The raw values.
            loc (str): The location, for warning messages.
            exclude_comments (bool): Whether to exclude the "comments"
                field.
                Default is False.
            is_notes (bool): Whether this dict is for "notes".
                Default is False.

        Returns:
            Tuple[bool, Dict]: Whether there was a warning,
                and the fixed dict.
        """

        fixed = self.make_default(is_notes)
        if exclude_comments:
            fixed.pop("comments", None)

        extra_bars = []
        missing_bars = {}
        unknown_fields = []
        missing_fields = {key: True for key in fixed.keys()}

        # special case: bars
        if "bars" in values:
            missing_fields.pop("bars")
            fixed_bars = fixed["bars"]
            missing_bars = {bar_num: True for bar_num in fixed_bars.keys()}
            for bar_num, val in values["bars"].items():
                if bar_num not in fixed_bars:
                    extra_bars.append(bar_num)
                    continue
                missing_bars.pop(bar_num)
                fixed_bars[bar_num] = val
        # save all fields except for name, link, and bars
        for key, val in values.items():
            if key in ("name", "link", "bars"):
                continue
            if key not in fixed:
                unknown_fields.append(key)
                continue
            missing_fields.pop(key)
            fixed[key] = val

        warning = False

        # warn about missing fields
        if len(missing_fields) > 0:
            warning = True
            logger.warning(
                "Missing fields {} for {}",
                ",".join(f'"{key}"' for key in missing_fields.keys()),
                loc,
            )
        if len(missing_bars) > 0:
            warning = True
            logger.warning(
                "Missing bar numbers {} for {}",
                ",".join(missing_bars.keys()),
                loc,
            )

        # warn about extra fields
        if len(extra_bars) > 0:
            warning = True
            logger.warning(
                "Extra bars {} for {} (not in range of 1-{})",
                ",".join(extra_bars),
                loc,
                self._bar_count,
            )
        if len(unknown_fields) > 0:
            warning = True
            logger.warning(
                "Unknown fields {} for {} "
                "(not in template definitions file)",
                ",".join(f'"{key}"' for key in unknown_fields),
                loc,
            )

        return warning, fixed

    def add_volunteer(self, email, data, strict=False):
        """Add a volunteer's data to this piece.

        Args:
            email (str): The volunteer email.
            data (Dict): The volunteer data.
            strict (bool): Whether to fail on warnings instead of only
                displaying them.
                Default is False.

        Returns:
            bool: Whether the addition was successful.
        """

        p_loc = f'volunteer "{email}", piece "{self._name}"'

        link = data["link"]
        warning = False
        if self._link is None:
            if link is not None:
                warning = True
                logger.warning('Extra piece link "{}" for {}', link, p_loc)
        else:
            if link is None:
                warning = True
                logger.warning("Missing piece link for {}", p_loc)
            elif link != self._link:
                warning = True
                logger.warning('Incorrect piece link "{}" for {}', link, p_loc)
        if strict and warning:
            fail_on_warning()
            return False

        # sources
        for name, source in data["sources"].items():
            if name not in self._sources:
                logger.warning(
                    'Unknown source "{}" for {} '
                    "(not in piece definitions file)",
                    name,
                    p_loc,
                )
                if strict:
                    fail_on_warning()
                    return False
                continue
            self._sources[name].add_volunteer(email, source)

        # notes
        notes = data["notes"]
        # special case: bars
        for bar_num, val in notes["bars"].items():
            if val == "":
                continue
            self._notes["bars"][bar_num][email] = val
        for key, val in notes.items():
            if key == "bars":
                continue
            if val == "":
                continue
            self._notes[key][email] = val

        return True



"""
create_sheet.py
Create a volunteer spreadsheet.
"""

# ======================================================================

import click
from loguru import logger

from ._input_files import (
    read_piece_definitions,
    read_template,
    read_volunteer_definitions,
)
from ._output_files import (
    SI_SUMMARY_KEY,
    read_spreadsheets_index,
    write_spreadsheets_index,
)
from ._shared import fail_on_warning
from ._sheets import (
    add_permissions,
    add_temp_sheet,
    create_spreadsheet,
    get_valid_worksheet_title,
    gspread_auth,
    open_spreadsheet,
    share_spreadsheet,
)

# ======================================================================

__all__ = ("create_sheet",)

# ======================================================================

INVITATION_MESSAGE = (
    "This is an invitation for you to collaborate on a crowd-sourcing "
    "spreadsheet for a project run by Roseingrave. Thank you for accepting "
    "our invitation as a volunteer."
)

# ======================================================================


class VolunteerSpreadsheet:
    """Represents a volunteer's spreadsheet."""

    def __init__(self, volunteer, url=None, extend=False):
        self._volunteer = volunteer
        self._url = url
        self._spreadsheet = None
        # If False, the spreadsheet will be wiped first
        self._extend_spreadsheet = extend

    @property
    def volunteer(self):
        return self._volunteer

    @property
    def extend_spreadsheet(self):
        return self._extend_spreadsheet

    def add_created_spreadsheet(self, spreadsheet):
        """Adds the given spreadsheet for this volunteer."""
        self._url = spreadsheet.url
        self._spreadsheet = spreadsheet

    def get_spreadsheet(self, gc):
        """Gets this volunteer's spreadsheet, opening it if necessary."""
        if self._spreadsheet is None:
            if self._url is None:
                raise RuntimeError(
                    "missing spreadsheet url for volunteer "
                    f'"{self._volunteer.email}"'
                )
            success, spreadsheet = open_spreadsheet(gc, self._url)
            if not success:
                return success, None
            self._spreadsheet = spreadsheet
        return True, self._spreadsheet


# ======================================================================


def create_spreadsheets(
    gc, spreadsheets, volunteer_spreadsheets, template, emails
):
    """Create spreadsheets for the given volunteers.

    Args:
        gc (gspread.Client): The client.
        spreadsheets (Dict[str, str]): The existing spreadsheet urls.
        volunteer_spreadsheets (Dict[str, VolunteerSpreadsheet]):
            The VolunteerSpreadsheet objects for each volunteer.
        template (Dict): The template settings.
        emails (Iterable[str]): The emails of the volunteers to create
            spreadsheets for.

    Returns:
        bool: Whether the creations were successful.
    """

    if len(emails) == 0:
        return True

    logger.info("Creating new spreadsheets")

    created = []

    def delete_created():
        for ss in created:
            gc.del_spreadsheet(ss)

    ss_settings = template["volunteerSpreadsheet"]
    folder = ss_settings["folder"]
    title_fmt = ss_settings["title"]
    public_access = ss_settings["publicAccess"]
    share_with_volunteer = ss_settings["shareWithVolunteer"]
    share_with = ss_settings["shareWith"]

    for email in emails:
        success, spreadsheet = create_spreadsheet(
            gc, title_fmt.format(email=email), folder
        )
        if not success:
            delete_created()
            return False
        created.append(spreadsheet.id)

        success = add_permissions(spreadsheet, public_access, share_with)
        if not success:
            delete_created()
            return False

        if share_with_volunteer:
            # make volunteer an editor
            success = share_spreadsheet(
                spreadsheet, email, "edit", notify=True, msg=INVITATION_MESSAGE
            )
            if not success:
                delete_created()
                return False

        spreadsheets[email] = spreadsheet.url
        volunteer_spreadsheets[email].add_created_spreadsheet(spreadsheet)

    return True


# ======================================================================


def populate_spreadsheets(gc, volunteer_spreadsheets, pieces, strict):
    """Populate the spreadsheets with the volunteer pieces.

    Args:
        gc (gspread.Client): The client.
        volunteer_spreadsheets (Dict[str, VolunteerSpreadsheet]):
            The VolunteerSpreadsheet objects for each volunteer.
        pieces (Dict[str, Piece]): The pieces.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.

    Returns:
        bool: Whether the population was successful.
    """

    logger.info("Populating spreadsheets with pieces")

    # piece name -> sheet object
    sheets = {}

    for email, volunteer_spreadsheet in volunteer_spreadsheets.items():
        logger.debug('Working on volunteer "{}"', email)
        volunteer = volunteer_spreadsheet.volunteer

        success, spreadsheet = volunteer_spreadsheet.get_spreadsheet(gc)
        if not success:
            if strict:
                fail_on_warning()
                return False
            continue

        existing_sheets = {
            sheet.title: sheet for sheet in spreadsheet.worksheets()
        }
        temp_sheet = None

        piece_names_set = set(volunteer.pieces)
        if volunteer_spreadsheet.extend_spreadsheet:
            # extend the sheet: only create the missing pieces
            # check for extra sheets that don't match the piece names
            extra_sheets = set(existing_sheets.keys()) - piece_names_set
            if len(extra_sheets) > 0:
                logger.warning(
                    "Found extra piece sheets: {}",
                    ",".join(f'"{sheet_name}"' for sheet_name in extra_sheets),
                )
                if strict:
                    fail_on_warning()
                    return False
        else:
            # delete existing sheets
            if len(existing_sheets) == 1:
                # no need to add temp sheet: use the only sheet
                _, temp_sheet = existing_sheets.popitem()
                # set the title of the temp sheet to something else, so
                # it doesn't conflict with the pieces being added
                temp_title = get_valid_worksheet_title(
                    spreadsheet, invalid=piece_names_set
                )
                temp_sheet.update_title(temp_title)
            else:
                # add a temp sheet
                success, temp_sheet = add_temp_sheet(
                    spreadsheet, invalid=piece_names_set
                )
                if not success:
                    if strict:
                        fail_on_warning()
                        return False
                    continue
                for sheet in existing_sheets.values():
                    spreadsheet.del_worksheet(sheet)
            existing_sheets.clear()

        # create the piece sheets
        worksheets_in_piece_order = []
        for piece_name in volunteer.pieces:
            if piece_name in existing_sheets:
                worksheets_in_piece_order.append(existing_sheets[piece_name])
                logger.debug(
                    'Skipping piece "{}": already exists in sheet', piece_name
                )
                continue
            if piece_name not in sheets:
                logger.debug('Creating sheet for piece "{}"', piece_name)
                sheet = pieces[piece_name].create_sheet(spreadsheet)
                sheets[piece_name] = sheet
            else:
                # copy from previously created
                logger.debug('Copying sheet for piece "{}"', piece_name)
                data = sheets[piece_name].copy_to(spreadsheet.id)
                # update sheet title to piece name
                sheet = spreadsheet.get_worksheet_by_id(data["sheetId"])
                sheet.update_title(piece_name)
            worksheets_in_piece_order.append(sheet)

        # delete temp sheet
        if temp_sheet is not None:
            spreadsheet.del_worksheet(temp_sheet)

        # reorder sheets according to piece order
        spreadsheet.reorder_worksheets(worksheets_in_piece_order)

    return True


# ======================================================================


@click.command(
    "create_sheet",
    help="Create volunteer spreadsheets.",
)
@click.argument("emails", type=str, nargs=-1)
@click.option(
    "-e",
    "--extend",
    is_flag=True,
    default=False,
    flag_value=True,
    help=(
        "Extend existing sheets with missing pieces. "
        "Cannot be set when `--replace` or `--new` is set."
    ),
)
@click.option(
    "-r",
    "--replace",
    is_flag=True,
    default=False,
    flag_value=True,
    help=(
        "Wipe and replace existing volunteer spreadsheets. "
        "Cannot be set when `--extend` is set. "
        "Has no effect when `--new` is set."
    ),
)
@click.option(
    "-n",
    "--new",
    is_flag=True,
    default=False,
    flag_value=True,
    help=(
        "Create new spreadsheets for all volunteers. "
        "Cannot be set when `--extend` is set."
    ),
)
@click.option(
    "-td",
    type=str,
    help="A filepath to replace the template definitions file.",
)
@click.option(
    "-pd", type=str, help="A filepath to replace the piece definitions file."
)
@click.option(
    "-vd",
    type=str,
    help="A filepath to replace the volunteer definitions file.",
)
@click.option(
    "-si", type=str, help="A filepath to replace the spreadsheets index file."
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    flag_value=True,
    help="Fail on warnings instead of only displaying them.",
)
def create_sheet(emails, extend, replace, new, td, pd, vd, si, strict):
    """Create volunteer spreadsheets.

    Args:
        emails (Tuple[str, ...]): The volunteers to create spreadsheets
            for. If none given, creates spreadsheets for all volunteers.
        extend (bool): Whether to extend existing sheets with missing
            pieces. New sheets will still be created.
            Cannot be True when `replace` or `new` is True.
            Default is False.
        replace (bool): Whether to wipe and replace existing volunteer
            spreadsheets.
            Cannot be True when `extend` is True.
            Has no effect when `new` is True.
            Default is False.
        new (bool): Whether to create new spreadsheets for all
            volunteers.
            Cannot be True when `extend` is True.
            Default is False.
        td (str): A filepath to replace the template definitions file.
        pd (str): A filepath to replace the piece definitions file.
        vd (str): A filepath to replace the volunteer definitions file.
        si (str): A filepath to replace the spreadsheets index file.
        strict (bool): Whether to fail on warnings instead of only
            displaying them.
            Default is False.
    """
    if extend and (replace or new):
        logger.error("`extend` can only be True on its own")
        return
    if replace and new:
        logger.warning("`--replace` has no effect since `--new` is set")

    success, gc = gspread_auth()
    if not success:
        return

    success, template = read_template(td, strict)
    if not success:
        return

    success, pieces = read_piece_definitions(template, pd)
    if not success:
        return

    success, volunteers = read_volunteer_definitions(pieces, vd, strict)
    if not success:
        return

    success, spreadsheets = read_spreadsheets_index(si)
    if not success:
        return

    if len(emails) > 0:
        # filter volunteers in `emails`
        filtered = {}
        email_not_found = False
        for email in emails:
            # don't allow this to update the summary spreadsheet
            # this technically won't change the actual behavior, since
            # there is no "SUMMARY" volunteer email
            if email == SI_SUMMARY_KEY:
                continue
            if email not in volunteers:
                email_not_found = True
                logger.warning(
                    'Volunteer "{}" not found in volunteer definitions file',
                    email,
                )
                continue
            filtered[email] = volunteers[email]
        if strict and email_not_found:
            fail_on_warning()
            return
        volunteers = filtered

    if new:
        # create new spreadsheets for all volunteers, even if they
        # already have a spreadsheet
        create_for = set(volunteers.keys())
    else:
        # don't create new for emails that already have a spreadsheet
        already_exist = {
            email for email in volunteers if email in spreadsheets
        }
        create_for = set(volunteers.keys()) - already_exist

        if extend or replace:
            # process the existing spreadsheets
            pass
        else:
            # remove existing spreadsheets: don't edit them
            for email in already_exist:
                logger.debug(
                    'Skipping volunteer "{}": '
                    "already in spreadsheets index file",
                    email,
                )
                volunteers.pop(email)

    if len(volunteers) == 0:
        logger.info("No volunteers to create spreadsheets for")
        return

    volunteer_spreadsheets = {}
    for email, volunteer in volunteers.items():
        if email in spreadsheets:
            # volunteer has existing spreadsheet
            url = spreadsheets[email]
            extend_spreadsheet = extend
        else:
            url = None
            extend_spreadsheet = False
        volunteer_spreadsheets[email] = VolunteerSpreadsheet(
            volunteer, url=url, extend=extend_spreadsheet
        )

    # create new spreadsheets
    success = create_spreadsheets(
        gc, spreadsheets, volunteer_spreadsheets, template, create_for
    )
    if not success:
        return

    # populate spreadsheets
    success = populate_spreadsheets(gc, volunteer_spreadsheets, pieces, strict)
    if not success:
        return

    if len(create_for) > 0:
        success = write_spreadsheets_index(spreadsheets, si)
        if not success:
            return

    logger.info("Done")



"""
_sheets.py
Shared methods for interacting with spreadsheets.
"""
# Need to access error data, but pylint complains
# Logs all unexpected exceptions
# pylint: disable=invalid-sequence-index,broad-except

# ======================================================================

import google.auth.exceptions
import gspread
from google_auth_oauthlib.flow import InstalledAppFlow
from loguru import logger

from ._read_write import get_path, read_settings
from ._shared import error

# ======================================================================

__all__ = (
    "gspread_auth",
    "create_spreadsheet",
    "open_spreadsheet",
    "get_valid_worksheet_title",
    "add_temp_sheet",
    "share_spreadsheet",
    "share_public",
    "add_permissions",
)

# ======================================================================


def _local_server_flow(client_config, scopes, port=0):
    """Runs an OAuth flow exactly like
    `gspread.auth.local_server_flow()`, except it does not force opening
    the user's browser.

    I don't like things that open browsers without warning, and a
    message will be displayed with the authentication URL, so it will
    still work.
    """
    flow = InstalledAppFlow.from_client_config(client_config, scopes)
    return flow.run_local_server(port=port, open_browser=False)


def gspread_auth(force=False):
    """Authenticate gspread to connect with Google Sheets.

    Args:
        force (bool): Whether to force the authentication.
            Default is False.

    Returns:
        Tuple[bool, gspread.Client]: Whether the setup was successful,
            and the client.
    """
    ERROR_RETURN = False, None

    def _error(msg):
        return error(msg, ERROR_RETURN)

    success = read_settings()
    if not success:
        return ERROR_RETURN

    logger.info("Setting up gspread authentication")
    try:
        filepath = get_path("credentials")
    except FileNotFoundError as ex:
        return _error(ex)

    auth_user_path = get_path("authorized_user", must_exist=False)
    # use `.resolve()` to turn this into the full path to avoid this:
    # https://github.com/burnash/gspread/issues/1056
    auth_user_path = auth_user_path.resolve()

    if force and auth_user_path.exists():
        auth_user_path.unlink()

    oauth_args = {
        "credentials_filename": str(filepath),
        "authorized_user_filename": str(auth_user_path),
        "flow": _local_server_flow,
        "client_factory": gspread.client.BackoffClient,
    }
    try:
        client = gspread.oauth(**oauth_args)
    except Exception as ex:
        # incorrect authorization code
        return _error(ex)

    # check if the OAuth Client is expired
    try:
        client.list_spreadsheet_files()
    except google.auth.exceptions.RefreshError:
        logger.warning("OAuth client credentials expired; refreshing")
        auth_user_path.unlink()
        # try again
        try:
            client = gspread.oauth(**oauth_args)
        except Exception as ex:
            # incorrect authorization code
            return _error(ex)

    return True, client


# ======================================================================


def create_spreadsheet(gc, name, folder=None):
    """Create a spreadsheet.

    Args:
        gc (gspread.Client): The client.
        name (str): The name of the spreadsheet.
        folder (Optional[str]): The id of the Google Drive folder to
            save the spreadsheet in.
            Default is None (save in root).

    Returns:
        Tuple[bool, gspread.Spreadsheet]:
            Whether the creation was successful, and the spreadsheet.
    """

    # code -> error reason -> message
    ERRORS = {
        404: {
            "notFound": (
                f'Couldn\'t find folder "{folder}". '
                "Please check that it's correct or remove it."
            ),
        },
    }

    try:
        return True, gc.create(name, folder)
    except gspread.exceptions.APIError as ex:
        args = ex.args[0]

        code = args["code"]
        if code in ERRORS:
            for err in args["errors"]:
                reason = err["reason"]
                if reason in ERRORS[code]:
                    error(ERRORS[code][reason])
                    return False, None

        # re-raise
        raise


# ======================================================================


def open_spreadsheet(gc, link):
    """Open a spreadsheet.

    Args:
        gc (gspread.Client): The client.
        link (str): The link of the spreadsheet to open.

    Returns:
        Tuple[bool, gspread.Spreadsheet]:
            Whether the open was successful, and the spreadsheet.
    """

    # code -> status -> message
    ERRORS = {
        403: {
            "PERMISSION_DENIED": (
                f'Couldn\'t open spreadsheet "{link}" (permission '
                "denied). Please make sure this spreadsheet is shared "
                "with you, or remove it from the spreadsheets "
                "index file."
            ),
        },
        404: {
            "NOT_FOUND": (
                f'Couldn\'t open spreadsheet "{link}" (not found). '
                "Please remove it from the spreadsheets index file."
            ),
        },
    }

    try:
        return True, gc.open_by_url(link)
    except gspread.exceptions.NoValidUrlKeyFound:
        error(f'Invalid spreadsheet link: "{link}"')
        return False, None
    except gspread.exceptions.APIError as ex:
        args = ex.args[0]

        code = args["code"]
        status = args["status"]

        try:
            logger.warning(ERRORS[code][status])
            return False, None
        except KeyError:
            pass

        # re-raise
        raise


# ======================================================================


def get_valid_worksheet_title(
    spreadsheet, title="_temp", fmt="{title}{num}", invalid=None
):
    """Returns a non-conflicting worksheet title.

    Args:
        spreadsheet (gspread.Spreadsheet): The spreadsheet.
        title (str): A title, to which a number will be appended until
            a valid worksheet title is found.
        fmt (str): A template format for how the number is appended to
            the title. Requires `"{title}"` and `"{num}"` to be included
            in the string.
            Default is `"{title}{num}"`.
        invalid (Set[str]): A set of invalid names.
            Default is None.

    Returns:
        str: A worksheet name.
    """
    if not ("{title}" in fmt and "{num}" in fmt):
        raise ValueError(
            '`fmt` is invalid: requires both "{title}" and "{num}"'
        )
    invalid_titles = set(invalid or set()) | set(
        worksheet.title for worksheet in spreadsheet.worksheets()
    )
    ws_title = title
    count = 1
    while ws_title in invalid_titles:
        ws_title = fmt.format(title=title, num=count)
        count += 1
    return ws_title


def add_temp_sheet(spreadsheet, invalid=None):
    """Add a temp sheet to a spreadsheet.

    Args:
        spreadsheet (gspread.Spreadsheet): The spreadsheet.
        invalid (Set[str]): A set of invalid names.
            Default is None.

    Returns:
        Tuple[bool, gspread.Worksheet]:
            Whether the addition was successful, and the temp sheet.
    """

    # code -> status -> message
    ERRORS = {
        403: {
            "PERMISSION_DENIED": (
                f'Couldn\'t edit spreadsheet "{spreadsheet.url}". '
                "Please make sure this spreadsheet is shared with you "
                "with edit permissions, or remove it from the "
                "spreadsheets index file."
            ),
        },
    }

    title = get_valid_worksheet_title(spreadsheet, invalid=invalid)

    try:
        return True, spreadsheet.add_worksheet(title, 1, 1)
    except gspread.exceptions.APIError as ex:
        args = ex.args[0]

        code = args["code"]
        status = args["status"]

        try:
            logger.warning(ERRORS[code][status])
            return False, None
        except KeyError:
            pass

        # re-raise
        raise


# ======================================================================


def _access_to_role(access, default):
    access = access.strip().lower()
    role = default
    if access == "view":
        role = "reader"
    elif access == "edit":
        role = "writer"
    return role


def _share(spreadsheet, email, *args, **kwargs):
    """Share a spreadsheet. Return True if successful."""

    # code -> error reason -> message
    ERRORS = {
        400: {
            "invalid": (f'Invalid email "{email}"'),
            "invalidSharingRequest": (
                # only happens if notify=False
                f'Invalid email "{email}": '
                "No Google account associated with this email address."
            ),
        },
        403: {
            "forbidden": (
                "Insufficient permissions to share spreadsheet "
                f'"{spreadsheet.title}".'
            )
        },
    }

    try:
        spreadsheet.share(email, *args, **kwargs)
        return True
    except gspread.exceptions.APIError as ex:
        args = ex.args[0]

        code = args["code"]
        if code in ERRORS:
            for err in args["errors"]:
                reason = err["reason"]
                if reason in ERRORS[code]:
                    error(ERRORS[code][reason])
                    return False

        # re-raise
        raise


def share_spreadsheet(
    spreadsheet, email, access="view", notify=False, msg=None
):
    """Share a spreadsheet with an email address.

    Args:
        spreadsheet (gspread.Spreadsheet): The spreadsheet.
        email (str): The email address.
        access (str): The access level.
            Options are "view" and "edit".
            Default is "view".
        notify (bool): Whether to notify the person.
            Default is False.
        msg (str): A message to send with the share notification.
            Default is None.

    Returns:
        bool: Whether the share was successful.
    """

    role = _access_to_role(access, "reader")

    success = _share(
        spreadsheet,
        email,
        perm_type="user",
        role=role,
        notify=notify,
        email_message=msg,
    )
    return success


def share_public(spreadsheet, access=None):
    """Share a spreadsheet publicly.
    Assumes the spreadsheet is not currently shared publicly.
    (That is, cannot restrict spreadsheet access; can only share.)

    Args:
        spreadsheet (gspread.Spreadsheet): The spreadsheet.
        access (Optional[str]): The access level.
            Choices are None, "view", and "edit".
            Default is None.

    Returns:
        bool: Whether the share was successful.
    """

    if access is None:
        return True

    role = _access_to_role(access, None)

    # don't share
    if role is None:
        return True

    success = _share(spreadsheet, None, perm_type="anyone", role=role)
    return success


def add_permissions(spreadsheet, public_access=None, share_with=None):
    """Add permissions to spreadsheet.

    Args:
        spreadsheet (gspread.Spreadsheet): The spreadsheet.
        public_access (Optional[str]): The public access level.
            Options are None, "view", and "edit".
            Default is None.
        share_with (List[str]): Email addresses to give edit access to.
            Default is None.

    Returns:
        bool: Whether the shares are successful.
    """
    if share_with is None:
        share_with = []

    try:
        success = share_public(spreadsheet, public_access)
        if not success:
            return False

        for email in share_with:
            success = share_spreadsheet(spreadsheet, email, "edit")
            if not success:
                return False
    except Exception as ex:
        error(ex)
        return False

    return True



"""
_read_write.py
Read and write operations on files.
"""

# ======================================================================

import json
from pathlib import Path

from loguru import logger

from ._shared import error

# ======================================================================

__all__ = (
    "get_path",
    "read_json",
    "write_json",
    "write_json_file",
    "read_settings",
    "fix_settings",
)

# ======================================================================

SETTINGS_FILE = "roseingrave.json"

# every time the script is run, this gets populated for that run
# becomes flattened version of the file
SETTINGS = {}

# ======================================================================


def _read_default(file):
    """Read a default configuration from the "defaults" directory."""
    filepath = Path(Path(__file__).parent, "defaults", file)
    contents = filepath.read_text(encoding="utf-8")
    return json.loads(contents)


# ======================================================================


def get_path(key, path=None, must_exist=True, create_dirs=False):
    """Get the path for a file key.

    Args:
        key (str): The key representing the file.
        path (Optional[str]): A path to the file to use instead.
            Default is None (use the settings file).
        must_exist (bool): Whether the file must exist.
            Default is True.
        create_dirs (bool): Whether to create missing parent
            directories.
            Default is False.

    Raises:
        ValueError: If `path` is None and the key is invalid.
        ValueError: If the file does not have extension ".json".
        FileNotFoundError: If `must_exist` is True
            and the file is not found.

    Returns:
        Path: The path to the file.
    """

    if path is None:
        if key not in SETTINGS:
            raise ValueError(f'invalid file key "{key}"')
        path = SETTINGS[key]
    else:
        path = [path]
    if not path[-1].endswith(".json"):
        raise ValueError('path must have extension ".json"')

    filepath = Path(*path)
    if must_exist and not filepath.exists():
        raise FileNotFoundError(f'file "{filepath}" not found')
    if create_dirs:
        filepath.parent.mkdir(parents=True, exist_ok=True)
    return filepath


# ======================================================================


def read_json(key, path=None):
    """Read data from a JSON file.

    Args:
        key (str): The key representing the file.
        path (Optional[str]): A path to the file to use instead.
            Default is None (use the settings file).

    Raises:
        ValueError: If `path` is None and the key is invalid.
        FileNotFoundError: If the file is not found.

    Returns:
        Union[Dict, List]: The contents of the JSON file.
    """
    filepath = get_path(key, path)
    contents = filepath.read_text(encoding="utf-8")
    return json.loads(contents)


def _write(filepath, data, msg):
    if msg is not None:
        logger.info('Writing {} to "{}"', msg, filepath)
    filepath.write_text(json.dumps(data, indent=2), encoding="utf-8")


def write_json(key, data, path=None, msg=None):
    """Write data to a JSON file.
    Creates the file and any parent directories.
    Overwrites file if already exists.

    Args:
        key (str): The key representing the file.
        data (Dict): The data to write.
        path (Optional[str]): A path to the file to use instead.
            Default is None (use the settings file).
        msg (Optional[str]): A message to log.
            If None, nothing is displayed.
            Default is None.

    Raises:
        ValueError: If `path` is None and the key is invalid.
    """
    filepath = get_path(key, path, must_exist=False, create_dirs=True)
    _write(filepath, data, msg)


def write_json_file(path, data, msg=None):
    """Write data to a JSON file.
    Creates the file and any parent directories.
    Overwrites file if already exists.

    Args:
        path (str): A path to the file.
        data (Dict): The data to write.
        msg (Optional[str]): A message to log.
            If None, nothing is displayed.
            Default is None.
    """
    filepath = Path(path)
    # make parent directories
    filepath.parent.mkdir(parents=True, exist_ok=True)
    _write(filepath, data, msg)


# ======================================================================


def _settings(msg, clear=False, must_exist=False):
    """Read the settings file and update the `SETTINGS` global variable.
    Returns whether successful and the fixed inputted settings.
    """
    ERROR_RETURN = False, None

    def _error(msg, *args, **kwargs):
        return error(msg.format(*args, **kwargs), ERROR_RETURN)

    if clear:
        SETTINGS.clear()
    if len(SETTINGS) > 0:
        return True, SETTINGS

    logger.info("{} settings file", msg)

    try:
        raw_values = read_json("", SETTINGS_FILE)
    except FileNotFoundError:
        if must_exist:
            _error("No settings file found to fix")
            return ERROR_RETURN
        # no settings file, so just use all defaults
        raw_values = {}

    # add defaults if missing
    defaults = _read_default("roseingrave.json")
    fixed = {}

    # credentials
    if "credentials" in raw_values:
        fixed["credentials"] = raw_values["credentials"]
        path = raw_values["credentials"]
    else:
        path = defaults["credentials"]
    if isinstance(path, str):  # turn into list if a str
        path = [path]
    SETTINGS["credentials"] = path
    SETTINGS["authorized_user"] = path[:-1] + ["authorized_user.json"]

    for level, level_defaults in defaults.items():
        # already took care of credentials
        if level in ("credentials",):
            continue
        if level in raw_values:
            fixed[level] = {}
            fixed_level = fixed[level]
            values = raw_values[level]
        else:
            values = {}
        for k, default in level_defaults.items():
            if k in values:
                path = values[k]
                fixed_level[k] = path
            else:
                path = default
            if isinstance(path, str):  # turn into list if a str
                path = [path]
            SETTINGS[k] = path

    # validate values
    invalid = False
    for key, value in SETTINGS.items():
        if not value[-1].endswith(".json"):
            invalid = True
            _error('"{}" must be a ".json" file', key)
    for key, search in (
        ("volunteerDataPath", "{email}"),
        ("pieceDataPath", "{piece}"),
    ):
        count = sum(path_piece.count(search) for path_piece in SETTINGS[key])
        if count != 1:
            invalid = True
            _error('"{}" must have "{}" exactly once', key, search)
    if invalid:
        return ERROR_RETURN

    return True, fixed


def read_settings():
    """Read the settings file.

    Returns:
        bool: Whether the read was successful.
    """
    success, _ = _settings("Reading")
    return success


def fix_settings():
    """Fix the settings file.

    Returns:
        bool: Whether the fix was successful.
    """
    success, fixed = _settings("Fixing", clear=True, must_exist=True)
    if not success:
        return False

    write_json("", fixed, SETTINGS_FILE, msg="fixed settings file")

    return True
